/*
 *  rommon Simple Monitor
 * 
 *  Copyright (C) 1999-2004 by Dep. of Computer Science and Engineering
 *                   Tomakomai National College of Technology, JAPAN
 *  Copyright (C) 2015-2019 by Roi Takeuchi
 *
 *  嵭Ԥϡʲ (1)(4) ξ狼Free Software Foundation 
 *  ˤäƸɽƤ GNU General Public License  Version 2 ˵
 *  ҤƤ˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơŬѲǽ
 *  ޤơʤݾڤԤʤޤܥեȥѤˤľ
 *  ŪޤϴŪʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: $
 */

/*
 *	rommon Serial Functions
 */
#include "sys_defs.h"
#include <string.h>
#include "sil.h"
#include <stdio.h>

#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

#define VECT_TAB_OFFSET     0x00		/* Vector Table base offset field. 
                                		   This value must be a multiple of 0x200. */

#define HSE_TIMEOUT_VALUE   (100*1000)	/* 100 ms */
#define HSI_TIMEOUT_VALUE   (2*1000)	/* 2 ms (minimum Tick + 1) */
#define LSI_TIMEOUT_VALUE   (2*1000)	/* 2 ms (minimum Tick + 1) */
#define HSI48_TIMEOUT_VALUE (2*1000)	/* 2 ms (minimum Tick + 1) */
#define PLL_TIMEOUT_VALUE   (2*1000)	/* 2 ms (minimum Tick + 1) */

#define CLOCKSWITCH_TIMEOUT_VALUE  (5000*1000) /* 5 s    */


#if !defined  (HSE_VALUE)
  #define HSE_VALUE         8000000	/*!< Value of the External oscillator in Hz */
#endif /* HSE_VALUE */

#if !defined  (HSI_VALUE)
  #define HSI_VALUE         16000000	/*!< Value of the Internal oscillator in Hz*/
#endif /* HSI_VALUE */

/* PLL_VCO = (HSE_VALUE or HSI_VALUE / PLL_M) * PLL_N */
#define PLL_M               4
#define PLL_N               85

/* SYSCLK = PLL_VCO / PLL_P */
#define PLL_P               2

/* USB FS, SDIO and RNG Clock =  PLL_VCO / PLLQ */
#define PLL_Q               2

/* I2SCLK =  PLLVCO / PLLR */
#define PLL_R               2	/* Possible value between 2 and 7 */

#define PLLM_SHIFT          4
#define PLLN_SHIFT          8
#define PLLR_SHIFT          25
#define PPRE1_SHIFT         8
#define PPRE2_SHIFT         11

#define PLL_MASK            (RCC_PLLCFGR_PLLSRC | RCC_PLLCFGR_PLLM | RCC_PLLCFGR_PLLN | \
                              RCC_PLLCFGR_PLLQ | RCC_PLLCFGR_PLLR | RCC_PLLCFGR_PLLPDIV)
#define PLL_VALUE           ((RCC_PLLSOURCE_HSI) | \
                             ((PLL_M - 1) << PLLM_SHIFT) | \
                             (PLL_N << PLLN_SHIFT) | \
                             (((PLL_Q >> 1) - 1) << 21) | \
                             (((PLL_R >> 1) - 1) << PLLR_SHIFT) | \
                             (PLL_P << 27))


/*
 *  ƥ९å
 */
#define RCC_SYSCLKSOURCE_HSI      RCC_CFGR_SW_HSI		/* HSI selection as system clock */
#define RCC_SYSCLKSOURCE_HSE      RCC_CFGR_SW_HSE		/* HSE selection as system clock */
#define RCC_SYSCLKSOURCE_PLLCLK   RCC_CFGR_SW_PLL		/* PLL selection as system clock */

#define RCC_PLL_SYSCLK            RCC_PLLCFGR_PLLREN	/* PLLCLK selection from main PLL */
#define RCC_HSICALIBRATION_DEFAULT 0x40	/* Default HSI calibration trimming value */

/*
 * FLASH_Latency FLASH Latency
 */ 
#define FLASH_LATENCY_0           FLASH_ACR_LATENCY_0WS		/* FLASH Zero wait state */
#define FLASH_LATENCY_1           FLASH_ACR_LATENCY_1WS		/* FLASH One wait state */
#define FLASH_LATENCY_2           FLASH_ACR_LATENCY_2WS		/* FLASH Two wait states */
#define FLASH_LATENCY_3           FLASH_ACR_LATENCY_3WS		/* FLASH Three wait states */
#define FLASH_LATENCY_4           FLASH_ACR_LATENCY_4WS		/* FLASH Four wait states */
#define FLASH_LATENCY_5           FLASH_ACR_LATENCY_5WS		/* FLASH Five wait state */
#define FLASH_LATENCY_6           FLASH_ACR_LATENCY_6WS		/* FLASH Six wait state */
#define FLASH_LATENCY_7           FLASH_ACR_LATENCY_7WS		/* FLASH Seven wait states */
#define FLASH_LATENCY_8           FLASH_ACR_LATENCY_8WS		/* FLASH Eight wait states */
#define FLASH_LATENCY_9           FLASH_ACR_LATENCY_9WS		/* FLASH Nine wait states */
#define FLASH_LATENCY_10          FLASH_ACR_LATENCY_10WS	/* FLASH Ten wait state */
#define FLASH_LATENCY_11          FLASH_ACR_LATENCY_11WS	/* FLASH Eleven wait state */
#define FLASH_LATENCY_12          FLASH_ACR_LATENCY_12WS	/* FLASH Twelve wait states */
#define FLASH_LATENCY_13          FLASH_ACR_LATENCY_13WS	/* FLASH Thirteen wait states */
#define FLASH_LATENCY_14          FLASH_ACR_LATENCY_14WS	/* FLASH Fourteen wait states */
#define FLASH_LATENCY_15          FLASH_ACR_LATENCY_15WS	/* FLASH Fifteen wait states */

#define TARGET_FLASH_LATENCY      FLASH_LATENCY_8
#define TARGET_HCLK_DIV           RCC_CFGR_HPRE_DIV1

/*
 *  PLL
 */
#define RCC_PLLSOURCE_NONE        0x00000000				/* No clock selected as PLL entry clock source  */
#define RCC_PLLSOURCE_HSI         RCC_PLLCFGR_PLLSRC_HSI	/* HSI clock selected as PLL entry clock source */
#define RCC_PLLSOURCE_HSE         RCC_PLLCFGR_PLLSRC_HSE	/* HSE clock selected as PLL entry clock source */

#define PWR_FLAG_SETTING_DELAY_US 50	/* Time out value for REGLPF and VOSF flags setting */


#define PWR_REGULATOR_VOLTAGE_SCALE1   PWR_CR1_VOS_0	/* Voltage scaling range 1 normal mode */
#define PWR_REGULATOR_VOLTAGE_SCALE2   PWR_CR1_VOS_1	/* Voltage scaling range 2             */

#define RCC_LPUART1CLKSOURCE_PCLK1     0x00000000U
#define RCC_LPUART1CLKSOURCE_SYSCLK    RCC_CCIPR_LPUART1SEL_0
#define RCC_LPUART1CLKSOURCE_HSI       RCC_CCIPR_LPUART1SEL_1
#define RCC_LPUART1CLKSOURCE_LSE       (RCC_CCIPR_LPUART1SEL_0 | RCC_CCIPR_LPUART1SEL_1)


/*
 *  CRS ErrorLimitDefault
 */
#define RCC_CRS_ERRORLIMIT_DEFAULT     0x00000022			/* Default Frequency error limit */

/*
 *  CRS HSI48֥졼
 */
#define RCC_CRS_HSI48CALIB_DEFAULT     0x00000020			/* Default Calibration HSI48 value */

/*
 *  CRS
 */
#define RCC_CRS_SYNC_SOURCE_GPIO  0x00000000				/* Synchro Signal source GPIO */
#define RCC_CRS_SYNC_SOURCE_LSE   CRS_CFGR_SYNCSRC_0		/* Synchro Signal source LSE */
#define RCC_CRS_SYNC_SOURCE_USB   CRS_CFGR_SYNCSRC_1		/* Synchro Signal source USB SOF (default)*/

/*
 *  CRS DIVIDER
 */
#define RCC_CRS_SYNC_DIV1         0x00000000				/* Synchro Signal not divided (default) */
#define RCC_CRS_SYNC_DIV2         CRS_CFGR_SYNCDIV_0		/* Synchro Signal divided by 2 */
#define RCC_CRS_SYNC_DIV4         CRS_CFGR_SYNCDIV_1		/* Synchro Signal divided by 4 */
#define RCC_CRS_SYNC_DIV8         (CRS_CFGR_SYNCDIV_1 | CRS_CFGR_SYNCDIV_0)	/* Synchro Signal divided by 8 */
#define RCC_CRS_SYNC_DIV16        CRS_CFGR_SYNCDIV_2		/* Synchro Signal divided by 16 */
#define RCC_CRS_SYNC_DIV32        (CRS_CFGR_SYNCDIV_2 | CRS_CFGR_SYNCDIV_0)	/* Synchro Signal divided by 32 */
#define RCC_CRS_SYNC_DIV64        (CRS_CFGR_SYNCDIV_2 | CRS_CFGR_SYNCDIV_1)	/* Synchro Signal divided by 64 */
#define RCC_CRS_SYNC_DIV128       CRS_CFGR_SYNCDIV			/* Synchro Signal divided by 128 */


#define get_sysclk_source()            (sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR)) & RCC_CFGR_SWS)
#define get_pll_source()               (sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR)) & RCC_PLLCFGR_PLLSRC)
#define set_hsi_calibration_adjust(a)  (sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_ICSCR), RCC_ICSCR_HSITRIM, (a) << 24))
#define get_flash_latency()            (sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY)
#define set_flash_latency(v)           (sil_modw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_LATENCY, (v)))

extern int MonVector[];

const unsigned char AHBPrescTable[16] = {0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 3, 4, 6, 7, 8, 9};
const unsigned char APBPrescTable[8] =  {0, 0, 0, 0, 1, 2, 3, 4};
uint32_t SystemCoreClock;


void
SetVector(void *addr)
{
	sil_wrw_mem((uint32_t *)(TADR_SCB_BASE+TOFF_SCB_VTOR), (unsigned int)addr | VECT_TAB_OFFSET);	/* Vector Table Relocation */
}


/*
 *  祷ƥ९å
 *  ͤϰʲ̤
 *      System Clock source            = PLL (HSI)
 *            SYSCLK(Hz)               = 170000000
 *            HCLK(Hz)                 = 170000000
 *            AHB Prescaler            = 1
 *            APB1 Prescaler           = 1
 *            APB2 Prescaler           = 1
 *            HSI Frequency(Hz)        = 16000000
 *            PLL_M                    = 4
 *            PLL_N                    = 85
 *            PLL_P                    = 2
 *            PLL_Q                    = 2
 *            PLL_R                    = 2
 *            Flash Latency(WS)        = 8
 *  return  ｪλʤ0֤
 */
int
SetSysClock(void)
{
	uint32_t tick = 0;
	uint32_t pllvco, pllsource, pllr, pllm;
	uint32_t pllfreq, pllcfgr, tmp;
	uint32_t wait_loop_index;

	/*
	 *  ѥ쥮졼BOOST⡼ɤ
	*/
	if((sil_rew_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1)) & PWR_CR1_VOS) == PWR_REGULATOR_VOLTAGE_SCALE2){
		/*
		 *  ߤVOLTAGE SCALE RANG2ξ硢RANG1-BOOSTͭ
		 */
		sil_andw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR5), PWR_CR5_R1MODE);

		/*
		 *  VOLTAGE SCALE RANG1
		 */
		sil_modw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR1), PWR_CR1_VOS, PWR_REGULATOR_VOLTAGE_SCALE1);

		/*
		 *  VOSFΥꥢԤ
		 */
		wait_loop_index = ((PWR_FLAG_SETTING_DELAY_US * SystemCoreClock) / 1000000U) + 1U;
		while((sil_rew_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_SR2)) & PWR_SR2_VOSF) != 0 && (wait_loop_index != 0U)){
			wait_loop_index--;
		}
		if((sil_rew_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_SR2)) & PWR_SR2_VOSF) != 0){
			return -1;
		}
	}
	else{
		/*
		 *  ߤVOLTAGE SCALE RANG1ξ硢RANG1-BOOSTͭΤߤԤ
		 */
		sil_andw_mem((uint32_t *)(TADR_PWR_BASE+TOFF_PWR_CR5), PWR_CR5_R1MODE);
    }

	/*
	 *  HSIåεư
	 */
    tmp = get_sysclk_source();
    pllsource = get_pll_source();
    if((tmp == RCC_CFGR_SWS_PLL && pllsource == RCC_PLLSOURCE_HSI) || tmp == RCC_CFGR_SWS_HSI){
		/*
		 *  HSIåưξ
		 */
        set_hsi_calibration_adjust(RCC_HSICALIBRATION_DEFAULT);

	}
	else{
		/*
		 *  HSIå̤ưξ
		 */
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_HSION);

		/*
		 *  HSIåΥǥԤ
		 */
        tick = 0;
		while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_HSIRDY) == 0){
			if(++tick > HSI_TIMEOUT_VALUE)
				return -1;
			 sil_dly_nse(1000);
		}

		/*
		 *  HSI֥졼
		 */
        set_hsi_calibration_adjust(RCC_HSICALIBRATION_DEFAULT);
    }

#ifdef USE_USB
	/*
	 *  HSI48ͭ
	 */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CRRCR), RCC_CRRCR_HSI48ON);

	/*
	 *  HSI48ǥԤ
	 */
	tick = 0;
	while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CRRCR)) & RCC_CRRCR_HSI48RDY) == 0){
		if(++tick > HSI48_TIMEOUT_VALUE)
			return -1;
		sil_dly_nse(1000);
	}

	/*
	 *  CRSե졼
	 */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1), RCC_APB1ENR1_CRSEN);
	tmp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1ENR1));
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1RSTR1), RCC_APB1RSTR1_CRSRST);
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APB1RSTR1), RCC_APB1RSTR1_CRSRST);

	/*
	 *  HSI48CRS
	 */
	tmp = (RCC_CRS_SYNC_DIV1 | RCC_CRS_SYNC_SOURCE_USB);
	tmp |= (48000000 / 1000) - 1;
	tmp |= (RCC_CRS_ERRORLIMIT_DEFAULT << 16);
	sil_wrw_mem((uint32_t *)(TADR_CRS_BASE+TOFF_CRS_CFGR), tmp);

	/*
	 *  ꥭ֥졼ȥߥ
	 */
	sil_modw_mem((uint32_t *)(TADR_CRS_BASE+TOFF_CRS_CR), CRS_CR_TRIM, (RCC_CRS_HSI48CALIB_DEFAULT << 8));

	/*
	 *  CRSͭ
	 */
	sil_orw_mem((uint32_t *)(TADR_CRS_BASE+TOFF_CRS_CR), (CRS_CR_AUTOTRIMEN | CRS_CR_CEN));

#endif

	/*
	 *  PLL(ƥॽPLLǤʤ)
	 */
	if(get_sysclk_source() != RCC_CFGR_SWS_PLL){
		/*
		 *  PLL
		 */
		sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_PLLON);

        /*
		 *  PLLԤ
		 */
        tick = 0;
		while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_PLLRDY) != 0){
			if(++tick > PLL_TIMEOUT_VALUE)
				return -1;
			sil_dly_nse(1000);
		}

        /*
		 *  PLLե奮졼
		 */
		sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR), PLL_MASK, PLL_VALUE);

		/*
		 *  PLLưȽ
		 */
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_PLLON);
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR), RCC_PLL_SYSCLK);

        /*
		 *  PLLǥԤ
		 */
        tick = 0;
		while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_PLLRDY) == 0){
			if(++tick > PLL_TIMEOUT_VALUE)
				return -1;
			sil_dly_nse(1000);
		}
	}
	else{
		/*
		 *  ղäʤΤǡͤǧ
		 */
		pllcfgr = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR));
		if((pllcfgr & PLL_MASK) != PLL_VALUE)
			return -1;
    }

	/*
	 *  ߤFLASH쥤ƥ󥷥åȤꡢϥåȤꤹ
	 */
	if(TARGET_FLASH_LATENCY > get_flash_latency()){
		set_flash_latency(TARGET_FLASH_LATENCY);

		/*
		 *  åȤѹ줿ƥå
		 */
		if(get_flash_latency() != TARGET_FLASH_LATENCY)
			return -1;
	}

	/*
	 *  SYSCLKPLL
	 *  PLL_VCO = (HSE_VALUE or HSI_VALUE/ PLLM) * PLLN
	 *  SYSCLK = PLL_VCO / PLLR
	 */
	pllsource = get_pll_source();
	pllcfgr = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR));
	pllm = ((pllcfgr & RCC_PLLCFGR_PLLM) >> PLLM_SHIFT) + 1;

	switch(pllsource){
	case RCC_PLLSOURCE_HSE:  /* HSE used as PLL clock source */
		pllvco = (HSE_VALUE / pllm) * ((pllcfgr & RCC_PLLCFGR_PLLN) >> PLLN_SHIFT);
		break;
	case RCC_PLLSOURCE_HSI:  /* HSI used as PLL clock source */
	default:
		pllvco = (HSI_VALUE / pllm) * ((pllcfgr & RCC_PLLCFGR_PLLN) >> PLLN_SHIFT);
		break;
	}
	pllr = (((pllcfgr & RCC_PLLCFGR_PLLR) >> PLLR_SHIFT) + 1) * 2;
	pllfreq = pllvco/pllr;

	/*
	 *  PLLå80MHz礭ϡDIV2Ȥꤢ
	 */
	if(pllfreq > 80000000U){
		sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_HPRE, RCC_CFGR_HPRE_DIV2);
	}

	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_SW, RCC_SYSCLKSOURCE_PLLCLK);

	tick = 0;
	while (get_sysclk_source() != (RCC_SYSCLKSOURCE_PLLCLK << 2)){
		if(++tick > CLOCKSWITCH_TIMEOUT_VALUE)
			return -1;
		sil_dly_nse(1000);
	}

	/*
	 *  HCLK
	 */
	/*  PLLȤ */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_PPRE1, RCC_CFGR_PPRE1_DIV16);
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_PPRE2, RCC_CFGR_PPRE2_DIV16);

    /*
	 *  HCLKΥǥХ
	 */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_HPRE, TARGET_HCLK_DIV);

	/*
	 *  ߤFLASH쥤ƥ󥷥åȤꡢ礭ϥåȤꤹ
	 */
	if (TARGET_FLASH_LATENCY < get_flash_latency()){
		set_flash_latency(TARGET_FLASH_LATENCY);

		/*
		 *  åȤѹ줿ƥå
		 */
		tick = 0;
		while(get_flash_latency() != TARGET_FLASH_LATENCY){
			if(++tick > CLOCKSWITCH_TIMEOUT_VALUE)
				return -1;
			sil_dly_nse(1000);
		}
	}

	/*
	 *  PCLK1,2
	 */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_PPRE1, RCC_CFGR_PPRE1_DIV1);
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_PPRE2, RCC_CFGR_PPRE2_DIV1);

	/*
	 *  LPUART1å
	 */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CCIPR), RCC_CCIPR_LPUART1SEL, RCC_LPUART1CLKSOURCE_PCLK1);
#if 0
  RCC_CRSInitTypeDef RCC_CRSInitStruct= {0};
  /*Enable CRS Clock*/
  __HAL_RCC_CRS_CLK_ENABLE();

  /* Default Synchro Signal division factor (not divided) */
//  RCC_CRSInitStruct.Prescaler = RCC_CRS_SYNC_DIV1;

  /* Set the SYNCSRC[1:0] bits according to CRS_Source value */
//  RCC_CRSInitStruct.Source = RCC_CRS_SYNC_SOURCE_USB;

  /* HSI48 is synchronized with USB SOF at 1KHz rate */
  RCC_CRSInitStruct.ReloadValue =  __HAL_RCC_CRS_RELOADVALUE_CALCULATE(48000000, 1000);
  RCC_CRSInitStruct.ErrorLimitValue = RCC_CRS_ERRORLIMIT_DEFAULT;

  /* Set the TRIM[5:0] to the default value */
  RCC_CRSInitStruct.HSI48CalibrationValue = RCC_CRS_HSI48CALIBRATION_DEFAULT;

  /* Start automatic synchronization */
  HAL_RCCEx_CRSConfig (&RCC_CRSInitStruct);

#endif
	return 0;
}

/*
 *  ߤSYSCLK, PCLK1, PCLK2ͤФؿ
 */
void
RCC_GetClocksFreq(RCC_ClocksTypeDef* RCC_Clocks)
{
	uint32_t pllvco, pllsource, pllr, pllm;
	uint32_t sysclockfreq, cfgr;
	uint32_t pllcfgr = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR));
	uint32_t syssource = get_sysclk_source();

	if(syssource == RCC_CFGR_SWS_HSI){
		/* HSI used as system clock source */
		sysclockfreq = HSI_VALUE;
	}
	else if(syssource == RCC_CFGR_SWS_HSE){
		/* HSE used as system clock source */
		sysclockfreq = HSE_VALUE;
	}
	else if(syssource == RCC_CFGR_SWS_PLL){
		/* PLL used as system clock  source */

		/* PLL_VCO = ((HSE_VALUE or HSI_VALUE)/ PLLM) * PLLN
		SYSCLK = PLL_VCO / PLLR
		*/
		pllsource = get_pll_source();
		pllm = ((pllcfgr & RCC_PLLCFGR_PLLM) >> PLLM_SHIFT) + 1;

		switch (pllsource){
		case RCC_PLLSOURCE_HSE:  /* HSE used as PLL clock source */
			pllvco = (HSE_VALUE / pllm) * ((pllcfgr & RCC_PLLCFGR_PLLN) >> PLLN_SHIFT);
			break;

		case RCC_PLLSOURCE_HSI:  /* HSI used as PLL clock source */
		default:
			pllvco = (HSI_VALUE / pllm) * ((pllcfgr & RCC_PLLCFGR_PLLN) >> PLLN_SHIFT);
			break;
		}
		pllr = (((pllcfgr & RCC_PLLCFGR_PLLR) >> PLLR_SHIFT) + 1) * 2;
		sysclockfreq = pllvco / pllr;
	}
	else{
		sysclockfreq = 0U;
	}

	cfgr = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR));
	RCC_Clocks->SYSCLK_Frequency = sysclockfreq;
	RCC_Clocks->PCLK1_Frequency = sysclockfreq >> (APBPrescTable[(cfgr & RCC_CFGR_PPRE1) >> PPRE1_SHIFT] & 0x1FU);
	RCC_Clocks->PCLK2_Frequency = sysclockfreq >> (APBPrescTable[(cfgr & RCC_CFGR_PPRE2) >> PPRE2_SHIFT] & 0x1FU);
}

/*
 *  ƥ
 */
void
SystemInit(void)
{
	RCC_ClocksTypeDef RCC_ClocksStatus;

	/*
	 *  FPU
	 */
	sil_orw_mem((uint32_t *)(TADR_SCB_BASE+TOFF_SCB_CPACR), ((3UL << 10*2)|(3UL << 11*2)));	/* set CP10 and CP11 Full Access */

	SystemCoreClock = 16000000;

	/*
	 *  PLL,ƥ९å
	 */
	if(SetSysClock() != 0){
		for(;;){}
	}
	RCC_GetClocksFreq(&RCC_ClocksStatus);
	SystemCoreClock = RCC_ClocksStatus.SYSCLK_Frequency;

	/*
	 *  VECTORơ֥
	 */
	SetVector((void*)MonVector);
}

